/**************************************************************
 *
 * $Header: /u/revow/cvs/src/knn/knnClass.c,v 1.2 1996/09/10 16:03:39 revow Exp $
 *
 * $Log: knnClass.c,v $
 * Revision 1.2  1996/09/10  16:03:39  revow
 * Added support for the different loss functions log(prob) and medain error
 *
 * Revision 1.1.1.1  1996/03/18  14:41:45  revow
 * CHecking in All version. Previous versions are under RCS control
 *
 * MODULE DESCRIPTION
 *
 *   Does K- nearest neighbour classification. The prediction is just the
 *   majority vote of the k-nearest neighbours. The option of picking 
 *   K using  Leave-One-Out Cross-validation is provided (looCv).
 *   Loss functions - K can be picked to minimize
 *   zero-one loss or square Prob loss. In the latter case a 
 *   predictive distributions for  class i is k_i / k. 
 *
 *   A few details have to be taken care of. These relate to situations
 *   where (a) training data points are close together in input space, and
 *   (b) Ties in the voting
 *
 *   To illustrate point (a), consider the case when  the k, k+1,
 *   k+p neighbours are within a tolerance of each other, we average
 *   these p training points to get the k'th point. (see notes and code 
 *   in evalPred)
 *
 *
 * PROJECT:     Delve
 *
 * $Author: revow $
 *
 *
 *************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <values.h>
#include <string.h>
#include "knnClass.h"


static char *optstring = "k:t:c:p:v:hm:r:";
static mrReal minProb = 1.e-06;

#define USAGE_STRING    \
"Usage %s : -[options] <test file> KNN for classification with a single output\nk - Set maximum K to try (default is num Training Examples)\nt - Training file (required)\nc - Name the prediction file for class prediction\np - Name the prediction file class probabilities\nr - Tolerance for gathering k'th neighbour default value is 1e-4 (see below)\nn - Set number of classes\nv - Set verbosity level (>= 0 )\nm Set method options\n  l - Do Leave-one-out CVV to pick best K\nh - Print this help\n\nWhen k, k+1, ..k+p points are very close, we replace these close proximity points\nby their mean. The r option sets the tolerance for choosing \'close\' proximity as:\n  (d2_{k+p} - d2_k)/ d2_k < tol.\nd2 is the squared distance\n"


int main(int argc, char *argv[])
{
    int		i;
    DESC   	desc;		/* Describes the problem */
    char        *tmplFile = "train";/* Template file */
    char	*testFile = "stdin";
    int         c;              /* Option character */
    extern char *optarg;        /* Next option argumnet */
    extern int  optind;         /* Index into next argument */
    FILE  	*fp_templ;
    FILE	*fp_test;
    int		nDim;
    int		testOut;

    (void)memset((void *)&desc, 0, sizeof(desc));
    desc.nClass = -1;
    desc.kVal = -1;
    desc.verbose = 1;
    desc.testOut = &testOut;
    desc.distTol = 1.e-4;
    desc.machTol = machineEpsilon();

    while ( (c = getopt(argc, argv, optstring)) != -1)
    {
      switch (c)
      {
       case 'h':
	fprintf(stderr, USAGE_STRING, argv[0]);
	exit (0);
	break;

       case 'k':
	desc.kVal = atoi(optarg);
	break;

       case 't':
	tmplFile = optarg;
	break;

       case 'c':
	desc.fpPred[L_ERR] = fopen(optarg, "w");
	break;

       case 'p':
	desc.fpPred[L_PROB] = fopen(optarg, "w");
	break;

       case 'v':
	desc.verbose = atoi(optarg);
	break;

       case 'r':
	desc.distTol = atof(optarg);
	break;

       case 'n':
	desc.nClass = atoi(optarg);
	break;

       case 'm':
	switch (*optarg)
	{
	 case 'l':
	  desc.methodOpts |= M_DO_LOO;
	  break;

	 default:
	  break;
	}
	break;

       default:
	break;
        }
    }

    if ( optind < argc)
    {
      testFile = argv[optind];
      fp_test = fopen(testFile, "r");
    }
    else
    {
      fp_test = stdin;
    }


    fp_templ = fopen(tmplFile, "r" );
    if ( fp_templ == NULL )
    {
	fprintf(stderr, "error in opening file %s\n", tmplFile);
	exit( 1 );
    }

    if (desc.verbose > 0)
    {
      fprintf(stdout, "%s: Training File: %s  Test File  %s\n", argv[0],
	      tmplFile, testFile);
    }
    countFileColsAndRows(tmplFile, fp_templ, &nDim, &desc.numTempl);
    desc.len_vec = nDim - 1;

    if (desc.nClass < 1)
    {
      desc.nClass = countClasses(fp_templ);
    }

    if (desc.len_vec < 1)
    {
      printf("Template vector lengths (%d)  must be > 0", desc.len_vec);
      exit(1);
    }


    if (desc.kVal < 1)
    {
      desc.kVal = desc.numTempl;
    }

    if (desc.nClass < 2)
    {
      fprintf(stderr, "Must have > 1 classes you have specified %d classes\n", desc.nClass);
      exit (1);
    }

    allocations(&desc);

    get_templates(fp_templ, &desc);
    
    
    if (desc.numTempl <= 0 || desc.kVal > desc.numTempl) {
	printf("kVal (%d) must not exceed totTempl (%d)\n", desc.kVal, 
	       desc.totTempl);
	exit(1);
    }
    
    /* get templates and test data */
    if (desc.methodOpts & M_DO_LOO)
    {
      looCv(&desc, fp_templ);
    }
    else
    {
      for (i = 0 ; i < NUM_LOSS ; ++i)
      {
	desc.bestK[i] = desc.kVal - 1;
      }
    }

    fclose(fp_templ);

    if (fp_test)
    {
      countFileColsAndRows(testFile, fp_test, &nDim, &desc.nTest);
      if ( nDim == desc.len_vec )
      {
	desc.testOut = NULL;
	if ( desc.verbose > 1 )
	{
	  fprintf(stdout, "Assuming DELVE type test data ie no Targets supplied\n");
	}
      }

      if (nDim == desc.len_vec || (nDim -1) == desc.len_vec)
      {
	int best = 0;
	/* Only need to evaluate up to best k */
	for (i = 0 ; i < NUM_LOSS ; ++i)
	{
	  best = MAX(best, desc.bestK[i]);
	}
	++best;
	/* desc.kVal = MIN(desc.kVal, best); */

       testExs(&desc, fp_test, FALSE);
      }
    }

    freeAllocations(&desc);
}


void allocations(
DESC		*desc		/* in: Problem description */
)
{
  int	i;

  desc->templates = (TEMPLATE *)calloc(desc->numTempl, sizeof(*desc->templates));

  desc->pred = (int *)calloc(desc->kVal, sizeof(*desc->pred));
  desc->testVec = (mrReal *)calloc(desc->len_vec+1, sizeof(*desc->testVec));

  for (i = 0 ; i < NUM_LOSS ; ++i)
  {
    desc->testErr[i] = (mrReal *)calloc(desc->kVal, sizeof(*desc->testErr));
  }

  desc->predProb = (mrReal **)malloc(desc->kVal * sizeof(*desc->predProb));

  for(i = 0 ; i < desc->kVal ; ++i)
  {
    desc->predProb[i] = (mrReal *)calloc(desc->nClass, sizeof(**desc->predProb));
  }
}

void freeAllocations(
DESC		*desc		/* in: Problem description */
)
{
  int	i;

  free((void *)desc->templates);
  free((void *)desc->pred);
  free((void *)desc->testVec);

  for (i = 0 ; i < NUM_LOSS ; ++i)
  {
    free((void *)desc->testErr[i]);
    if (desc->fpPred[i])
    {
      fclose(desc->fpPred[i]);
    }
  }

  for(i = 0 ; i < desc->kVal ; ++i)
  {
    /* free((void *)desc->predProb[i]); */
  }
  free((void *)desc->predProb);
}

void get_templates(
FILE		*fp,
DESC		*desc		/* in: Problem description */
)
{
  int		i;
  int		cont = 1;
  int		class;
  int		*pClass = NULL;
  mrReal	*vec;		/* Current template vector */
  int		val;

  i = 0;

  while (   (i < desc->numTempl)
	    && (vec = (mrReal *)malloc(desc->len_vec * sizeof(*vec)))
	    && (vec = readRow(fp, vec, desc->len_vec, &val)) )
  {
    desc->templates[i].vec = vec;
    desc->templates[i].out = val;
    desc->templates[i].id = i;
    ++i;
  }

  desc->numTempl = i;
}
/**************************************************************
 *
 *
 * NAME: testExs
 *
 * DESCRIPTION:
 *
 *  Perform knnn  neighbours on the data in the file pointed by fp
 *  To simulate leave-one-out CV can ask that the i'th training 
 *  example be left out (SET leavOUT = TRUE)
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
void testExs(
DESC		*desc,		/* in: Problem description */
FILE		*fp,		/* in: Where to read test data */
int		leaveOut
)
{
  int		i, j, l;
  mrReal	*v;
  int		*pred;
  mrReal	tmp, numExs;
  mrReal	mean, var;	/* Mean var for predictive gaussian */
  mrReal	sqPloss;	/* Squared Prob */

  for( i = 0 ; i < NUM_LOSS ; ++i)
  {
    for (j = 0 ; j < desc->kVal ; ++j)
    {
      desc->testErr[i][j] = 0.0;
    }
  }

  rewind(fp);
  v = desc->testVec;
  desc->leaveOut = -1;

  for(i = 0 ; readRow(fp, v, desc->len_vec, desc->testOut) != NULL  ; ++i) 
  {
    if (TRUE == leaveOut)
    {
      desc->leaveOut = MIN(i, desc->numTempl);
    }

    pred = classifyEx(desc);

    /*
     * Accumulate loss functions for all k
     */
    for (j = 0 ; desc->testOut && j < desc->kVal ; ++j)
    {
      if (pred[j] !=  *desc->testOut)
      {
	++desc->testErr[L_ERR][j];
      }

      sqPloss = 0.0;
      for (l = 0 ; l < desc->nClass ; ++l)
      {
	if (*desc->testOut == l)
	{
	  tmp = 1.0 - desc->predProb[j][l];
	  sqPloss += tmp * tmp;
	}
	else
	{
	  sqPloss += desc->predProb[j][l] * desc->predProb[j][l];
	}
      }

      desc->testErr[L_PROB][j] += sqPloss;
      if (j == desc->bestK[L_PROB])
      {
	desc->reportVal[L_PROB] = sqPloss;
      }

    }

    if (FALSE == leaveOut)
    {
      desc->reportVal[L_ERR] =  pred[desc->bestK[L_ERR]];

      for (j = 0 ; j < NUM_LOSS ; ++j)
      {
	if (desc->fpPred[j] && desc->bestK[j] >= 0)
	{
	  if (L_PROB == j)
	  {
	    for (l = 0 ; l < desc->nClass ; ++l)
	    {
	      fprintf(desc->fpPred[j], "%g ", desc->predProb[desc->bestK[L_PROB]][l]);
	    }
	    fprintf(desc->fpPred[j], "\n");
	  }
	  else
	  {
	    fprintf(desc->fpPred[j], "%g\n", desc->reportVal[j]);
	  }
	}
      }
    }

  }

  numExs = i;

  if (desc->verbose > 1)
  {
    fprintf(stdout, "Number of train Cases: %d  Test Cases: %d Input Dimension %d\n", desc->numTempl, i, desc->len_vec);


    if (desc->testOut && desc->verbose > 2)
    {
      printf("k   ERROR    PROB\n");

      for(i = 0 ; i < desc->kVal ; ++i)
      {
	fprintf(stdout, "%d   ", i+1);

	for (j = 0 ; j < NUM_LOSS ; ++j)
	{
	  fprintf(stdout, "%6.3g  ",  desc->testErr[j][i] / numExs);
	}

	fprintf(stdout, "\n");
      }
    }
  }

}
/**************************************************************
 *
 *
 * NAME: looCv
 *
 * DESCRIPTION:
 *
 *  Perform Leave One Out Cross validation
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
void looCv(
DESC		*desc,		/* in: Problem description */
FILE		*fp		/* in: Source of the training data */
)
{
  int		i, j;
  mrReal	 bestCv[NUM_LOSS];

  testExs(desc, fp, TRUE);

  for (i = 0 ; i < NUM_LOSS ; ++i)
  {
    desc->bestK[i] = 0;
    bestCv[i] = desc->testErr[i][0];

    for (j = 1 ; j < desc->kVal ; ++j)
    {
      if(desc->testErr[i][j] < bestCv[i])
      {
	bestCv[i] = desc->testErr[i][j];
	desc->bestK[i] = j;
      }
    }
  }

  if (desc->verbose > 0)
  {
    for (i = 0 ; i < NUM_LOSS ; ++i)
    {
       fprintf(stdout, "Leave-One-Out Best K (%s) = %d\n", lossDesc[i], desc->bestK[i] + 1);
    }
  }
}

/**************************************************************
 *
 *
 * NAME: classifyEx
 *
 * DESCRIPTION:
 *
 *  Classify an example
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
int * classifyEx(
DESC		*desc		/* in: Problem description */
)
{
  int		i, j, k;
  TEMPLATE	*template;
  int		tempNum = 0;

  j = k = 0;
  template = desc->templates;

    /*
     *   Compute the distances for each template class
     */
  for (i = 0 ; i < desc->numTempl ; ++i, template++)
  {
    if (template->id != desc->leaveOut)
    {
      calc_dist2_to_one(desc->testVec, template, desc->len_vec);
    }
    else
    {
      template->dist2 = MAXFLOAT;
    }
  }


  sort_dist2(desc->templates, desc->numTempl);
  ++desc->nTest;

  for (k = 1 ; k <= desc->kVal ; ++k)
  {
    desc->pred[k-1] = evalPred(desc, k);
  }

  return (desc->pred);
}

/**************************************************************
 *
 *
 * NAME: evalPred
 *
 * DESCRIPTION:
 *
 *  Evaluates the predictor based on the k NN. The best guess
 *  is simply the clas with the majority of K-NN. In the case where
 *  there are > 1 classes with the most NN, we pick the class
 *  which has the smallest summed squared distance of neighbours
 *  to the test point.
 *  We go to great lengths to avoid
 *  inconsistant results when the k-m, ... (k-1), k, (k+1),... (k+p)
 *  points are all very close (within tol) of each other. What we in effect
 *  say is that we want to get the same result regardless how these
 *  (p+m+1) are ordered (Normally we expect p=m=0). One way to do this
 *  is to average these (p+m+1) points as y_k = (1/(p+m+1) sum_i y_i
 *  where the sum is over these near points. Finally we compute
 *  the predictor as:
 *
 *    y = 1/k (sum_i=1..(k-m-1) y_i  + (m+1)/(m+p+1) sum_i=(k-m), ... (k+p) y_i
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
int evalPred(
DESC		*desc,		/* in: Problem description */
int		k		/* in: k-value */
)
{
  int		i, m, p, k1;
  mrReal	tmp;
  mrReal	d_k;		/* Distance to k-nn */
  mrReal	tol;
  int		best;
  mrReal	*classCnt;

  k1 = k - 1;
  d_k = desc->templates[k1].dist2;
  tol = d_k * desc->distTol;
  p = m = 0;
  classCnt = desc->predProb[k1];


  for (i = 0 ; i < desc->nClass ; ++i)
  {
    classCnt[i] = 0.0;
  }

  /*
   * First off evaluate m
   */
  for(i = k1 ; i >= 0 && (d_k - desc->templates[i].dist2) <= tol ; --i)
  {
    classCnt[desc->templates[i].out]++;
    ++m;
  }


  /*
   * Now evaluate the k+1,... k+p close neighbours
   */
  for( i=k; i < desc->kVal  && (desc->templates[i].dist2 - d_k) < tol ; ++i)
  {
    if (desc->templates[i].out < 0 || desc->templates[i].out >= desc->nClass)
    {
      fprintf(stderr, "ERROR!\n");
    }

    classCnt[desc->templates[i].out]++;
    ++p;
  }

  tmp = (mrReal)m / (m+p);
  for (i = 0 ; i < desc->nClass ; ++i)
  {
    classCnt[i] *= tmp;
  }


  for (i = 0 ; i < k - m ; ++i)
  {
    if (desc->templates[i].out < 0 || desc->templates[i].out >= desc->nClass)
    {
      fprintf(stderr, "ERROR!\n");
    }
    classCnt[desc->templates[i].out]++;
  }


  tmp = 1.0 / (mrReal)k;

  best = 0;
  for (i = 0 ; i < desc->nClass ; ++i)
  {
    classCnt[i] *= tmp;

    if (classCnt[i] > classCnt[best])
    {
      best = i;
    }
    else if (best != i && classCnt[i] > 0.0 && classCnt[i] == classCnt[best])
    {
      best = smallerDist(desc, k+p, i, best);
    }
  }

  return (best);
}

/**************************************************************
 *
 *
 * NAME: smallerDist
 *
 * DESCRIPTION:
 *
 *  Used to break ties when there are equal number of neighbours
 *  to the test point from two classes (m,n). We break the tie
 *  by choosing the class whose neighbours have the smallest
 *  summed distance to the test point.
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
int smallerDist(
DESC		*desc,	/* in: Problem description */
int		maxK,	/* in Max k to consider (includes p that may be 
			   within tol of each other (See evalPred above) */
int		m,	/* in: Class 1 */
int		n	/* Class 2 */
)
{
  int		i;
  mrReal	mdist = 0.0, ndist = 0.0;

  if (m >= desc->nClass || n >= desc->nClass)
  {
    fprintf(stderr, "Invalid classes in %d %d both should be < %d\n",
	    m, n, desc->nClass);
    exit(1);
  }

  for (i = 0 ; i < maxK ; ++i)
  {
    if (desc->templates[i].out == m)
    {
      mdist += desc->templates[i].dist2;
    }
  
    if (desc->templates[i].out == n)
    {
      ndist += desc->templates[i].dist2;
    }
  }

  return (mdist < ndist ? m : n);
}

void   calc_dist2_to_one(
mrReal			*tst,		/* in: Example Vector */
TEMPLATE		*templ,		/* out: Set the distance function */
int			len
)
{
  mrReal 		tmp;
  int 		i;

  templ->dist2 = 0.0;

  for (i = 0 ; i < len ; i++, ++tst)
  {
    tmp = templ->vec[i] - *tst;
    templ->dist2 += tmp * tmp;
  }

}

void  sort_dist2(
TEMPLATE 	*templates,
int		num
)
{
  qsort( (void *)templates, (size_t)num, sizeof(*templates), compare_templ);
}

int  compare_templ(
const void *arg1,
const void *arg2
)
{
  TEMPLATE  *first = (TEMPLATE *)arg1;
  TEMPLATE *second = (TEMPLATE *)arg2;
  mrReal d_first, d_second;	

  d_first = first->dist2;
  d_second = second->dist2;
  
  if ((d_first) == (d_second))
    return(0);
  else if ((d_first) > (d_second))
    return(1);
  else
    return(-1);

}

/**************************************************************
 *
 *
 * NAME: readRow
 *
 * DESCRIPTION:
 *
 *  Read a row into a vector
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  Vector read
 *
 *************************************************************/
mrReal *readRow(
FILE		*fp,		/* in: Where to read from */
mrReal		*vec,		/* i/o: where to put data */
int		num,		/* in: Number to read */
int		*out		/* out: Class of example */
)
{
    int		i;
    char	*buf;
    mrReal	*ret;		/* Returned Vector */

    ret = vec;

    for (i = 1 ; vec && i <= num ; ++i)
    {
      if (1 != fscanf(fp, "%f ", vec++))
      {
	vec = NULL;
	break;
      }
    }

    if (NULL == vec)
    {
	free((void *) ret);
	ret = NULL;
    }
    else if (out)
    {
	fscanf(fp, "%d", out);
    }

    return (ret);
}
/**************************************************************
 *
 *
 * NAME: countFileColsAndRows
 *
 * DESCRIPTION:
 *
 *  Counts number of columns in a file and number of
 *  lines
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 *  
 *
 *************************************************************/
int countFileColsAndRows(
char		*file,
FILE		*fp,		/* in: Data file */
int		*ncol,		/* out: # of cols */
int		*nlines		/* out Estimated # lines */
)
{
  char	line[BUFSIZ];
  char	*lp, *lp1;
  long	nc = 0;		/* Column count on current line */


    /*
     * Count the number of lines in the file
     */
  *nlines = 0;
  while (fgets(line, BUFSIZ, fp))
  {
    nc = countCols(line, BUFSIZ);

    if (*nlines > 0)
    {
      if (nc != *ncol)
      {
	fprintf(stderr, "%s: Inconsistant number of columns on line %d\n", file, *nlines);
	return(-1);
      }
    }   
    *ncol = nc;
    ++(*nlines);
  }

  rewind(fp);

}
/**************************************************************
 *
 *
 * NAME: countCols
 *
 * DESCRIPTION:
 *
 *  Counts columns of data in a string
 *
 * CAVEATES
 *
 *  
 * RETURNS:
 *
 * Number of columns 
 *
 *************************************************************/
static int countCols(
char		*line,		/* in: Line with data */
int		maxLen		/* in: Maximum line length */
)
{
  char	*lp, *lp1;
  int		ncol;

  lp1 =  line;
  lp = NULL;

  for (ncol = -1 ; lp != lp1 && (lp1 - line < maxLen) ; ++ncol)
  {
    lp = lp1;
    (void)strtod(lp, &lp1);
  }

  return (ncol);
}


mrReal machineEpsilon()
{
  mrReal x;
  double eps = 1.0;

  for (;;) 
  {
    x = 1 + eps/2.0;
    if (x>1.0)
    {
      eps /= 2.0;
    }
    else
    {
      break;
    }
  }

  return eps;
}
static int countClasses(
FILE	*fp
)
{
    char	line[BUFSIZ];
    char	*lp, *lp1;
    int		numClass= 0;
    int		*classList = NULL;
    float	tmp;
    int		i, class;

    while (fgets(line, BUFSIZ, fp))
    {
      lp1 =  line;
      lp = NULL;
      tmp = 0.0;

      for (i = -1 ; lp != lp1 ; ++i)
      {
	lp = lp1;
	class = tmp;

	tmp = strtod(lp, &lp1);
      }

      for (i = 0 ; i < numClass ; ++i)
      {
	if (class == classList[i])
	{
	  break;
	}
      }

      if (i >= numClass)
      {
	i = numClass+1;
	STATIC_BUFFER(classList, int, i, numClass);
	
	classList[i-1] = class;
      }

    }

    rewind(fp);
    return numClass;
}
